/*
	File:		KeyboardModule.c

	Contains:	HID Module for USB Keyboard

	Version:	xxx put version here xxx

	Copyright:	 1997-1998 by Apple Computer, Inc., all rights reserved.

*/

#include <Types.h>
#include <Devices.h>
#include <processes.h>
#include <DriverServices.h>
#include <USB.h>

#include "KeyboardModule.h"

usbKeyboardPBStruct myKeyboardPB;
usbKeyboardPBStruct shimKeyboardPB;



void InitParamBlock(USBReference theInterfaceRef, USBPB * paramblock)
{
	paramblock->usbReference = theInterfaceRef;
	paramblock->pbVersion = kUSBCurrentPBVersion;
	
	paramblock->usb.cntl.WIndex = 0; 			
	paramblock->usb.cntl.WValue = 0;
	
	paramblock->usbBuffer = nil;		
	paramblock->usbActCount = 0;
	paramblock->usbReqCount = 0;
	paramblock->usbFlags = 0;
	paramblock->usbOther = 0;
	
	paramblock->usbStatus = noErr;
}

static Boolean immediateError(OSStatus err)
{
	return((err != kUSBPending) && (err != noErr) );
}

void KeyboardInitiateTransaction(USBPB *pb)
{
register usbKeyboardPBStruct *pKeyboardPB;
OSStatus myErr;

	pKeyboardPB = (usbKeyboardPBStruct *)(pb);
	pKeyboardPB->transDepth++;
	if (pKeyboardPB->transDepth < 0)
	{
		USBExpertFatalError(pKeyboardPB->interfaceRef, kUSBInternalErr, "\pUSBHIDKeyboardModule: transDepth < 0 (initiation)", pKeyboardPB->pb.usbRefcon );
	}
	
	if (pKeyboardPB->transDepth > 1)
	{
		USBExpertFatalError(pKeyboardPB->interfaceRef, kUSBInternalErr, "\pUSBHIDKeyboardModule: transDepth > 1 (initiation)", pKeyboardPB->pb.usbRefcon );
	}
	
	switch(pKeyboardPB->pb.usbRefcon & ~kRetryTransaction)
	{
		case kSetKeyboardLEDs:
			InitParamBlock(pKeyboardPB->interfaceRef, &pKeyboardPB->pb);
			
			pKeyboardPB->pb.usb.cntl.BMRequestType = USBMakeBMRequestType(kUSBOut, kUSBClass, kUSBInterface);			
			
			pKeyboardPB->pb.usb.cntl.BRequest = kHIDRqSetReport;
			pKeyboardPB->pb.usb.cntl.WValue = (kHIDRtOutputReport << 8); 
			pKeyboardPB->pb.usb.cntl.WIndex = pKeyboardPB->interfaceDescriptor.interfaceNumber;
			
			pKeyboardPB->pb.usbBuffer = (Ptr)&pKeyboardPB->hidReport[0];
			pKeyboardPB->pb.usbReqCount = 1;
				
			pKeyboardPB->pb.usbCompletion = (USBCompletion)KeyboardCompletionProc;
			pKeyboardPB->pb.usbRefcon |= kCompletionPending;
			
			myErr = USBDeviceRequest(&pKeyboardPB->pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pKeyboardPB->interfaceRef, kUSBInternalErr, "\pUSBHIDKeyboardModule: kSetKeyboardLEDs - immediate error", myErr);
			}
			break;

		case kConfigureInterface:
			InitParamBlock(pKeyboardPB->interfaceRef, &pKeyboardPB->pb);
			
			pKeyboardPB->pb.usbCompletion = (USBCompletion)KeyboardCompletionProc;
			pKeyboardPB->pb.usbRefcon |= kCompletionPending;
			
			myErr = USBConfigureInterface( &pKeyboardPB->pb );
			if(immediateError(myErr))
			{
				USBExpertFatalError(pKeyboardPB->interfaceRef, kUSBInternalErr, "\pUSBHIDKeyboardModule: kConfigureInterface - immediate error", myErr);
				pKeyboardPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
		
		case kSetProtocol:
			InitParamBlock(pKeyboardPB->interfaceRef, &pKeyboardPB->pb);
			
			pKeyboardPB->pb.usb.cntl.BMRequestType = USBMakeBMRequestType(kUSBOut, kUSBClass, kUSBInterface);			
			pKeyboardPB->pb.usb.cntl.BRequest = kHIDRqSetProtocol;
			pKeyboardPB->pb.usb.cntl.WValue = kHIDBootProtocolValue; 
			pKeyboardPB->pb.usb.cntl.WIndex = pKeyboardPB->interfaceDescriptor.interfaceNumber;
			
			pKeyboardPB->pb.usbCompletion = (USBCompletion)KeyboardCompletionProc;
			pKeyboardPB->pb.usbRefcon |= kCompletionPending;
		
			myErr = USBDeviceRequest(&pKeyboardPB->pb);
			if (immediateError(myErr))
			{
				USBExpertFatalError(pKeyboardPB->interfaceRef, kUSBInternalErr, "\pUSBHIDKeyboardModule: kSetProtocol - immediate error", myErr);
			}
			break;
			
		case kSetIdleRequest:
			InitParamBlock(pKeyboardPB->interfaceRef, &pKeyboardPB->pb);
			
			pKeyboardPB->pb.usb.cntl.BMRequestType = USBMakeBMRequestType(kUSBOut, kUSBClass, kUSBInterface);			
			
			pKeyboardPB->pb.usb.cntl.BRequest = kHIDRqSetIdle;
			pKeyboardPB->pb.usb.cntl.WValue = ((24/4)<<8); 				// force a read completion if idle for more than 24ms
			pKeyboardPB->pb.usb.cntl.WIndex = pKeyboardPB->interfaceDescriptor.interfaceNumber;
			
			pKeyboardPB->pb.usbCompletion = (USBCompletion)KeyboardCompletionProc;
			pKeyboardPB->pb.usbRefcon |= kCompletionPending;

			myErr = USBDeviceRequest(&pKeyboardPB->pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pKeyboardPB->interfaceRef, kUSBInternalErr, "\pUSBHIDKeyboardModule: kSetIdleRequest - immediate error", myErr);
			}
			break;

		case kFindPipe:
			InitParamBlock(pKeyboardPB->interfaceRef, &pKeyboardPB->pb);
			
			pKeyboardPB->pb.usbFlags = kUSBIn;
			pKeyboardPB->pb.usbClassType = kUSBInterrupt;
			
			pKeyboardPB->pb.usbCompletion = (USBCompletion)KeyboardCompletionProc;
			pKeyboardPB->pb.usbRefcon |= kCompletionPending;
		
			myErr = USBFindNextPipe( &pKeyboardPB->pb );
			if((immediateError(myErr)) || (pKeyboardPB->pb.usbBuffer == nil))
			{
				USBExpertFatalError(pKeyboardPB->interfaceRef, kUSBInternalErr, "\pUSBHIDKeyboardModule: kFindPipe - immediate error", myErr);
				pKeyboardPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
		
		case kReadInterruptPipe:
			InitParamBlock(pKeyboardPB->pipeRef, &pKeyboardPB->pb);

			pKeyboardPB->pb.usbBuffer = (Ptr)pKeyboardPB->hidReport;
			pKeyboardPB->pb.usbReqCount = 0x08;
			pKeyboardPB->pb.usb.cntl.WIndex = pKeyboardPB->interfaceDescriptor.interfaceNumber;	
			
			pKeyboardPB->pb.usbCompletion = (USBCompletion)KeyboardCompletionProc;
			pKeyboardPB->pb.usbRefcon |= kCompletionPending;
		
			myErr = USBIntRead(&pKeyboardPB->pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pKeyboardPB->interfaceRef, kUSBInternalErr, "\pUSBHIDKeyboardModule: kReadInterruptPipe - immediate error", myErr);
			}
			break;
			
		case kClearFeature:
			USBExpertStatus(pKeyboardPB->interfaceRef, "\pUSBHIDKeyboardModule: Do a clear feature on the interrupt endpoint", pKeyboardPB->pipeRef);
			InitParamBlock(pKeyboardPB->pipeRef, &pKeyboardPB->pb);
			pKeyboardPB->pb.usb.cntl.BMRequestType = USBMakeBMRequestType(kUSBOut, kUSBStandard, kUSBEndpoint);
			
			pKeyboardPB->pb.usb.cntl.BRequest = kUSBRqClearFeature;
			pKeyboardPB->pb.usb.cntl.WValue = kUSBFeatureEndpointStall;
			
			pKeyboardPB->pb.usbFlags = kUSBAddressRequest;				/* kUSBAddressRequest asks the USL to do a translation of piperef to endpoint number */

			pKeyboardPB->pb.usbCompletion = (USBCompletion)KeyboardCompletionProc;
			pKeyboardPB->pb.usbRefcon |= kCompletionPending;

			myErr = USBDeviceRequest(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pKeyboardPB->pb.usbReference, kUSBInternalErr, "\pUSBHIDKeyboardModule: kClearFeature - immediate error", myErr);
			}
			break;
				
		default:
			USBExpertFatalError(pKeyboardPB->interfaceRef, kUSBInternalErr, "\pUSBHIDKeyboardModule - Transaction completed with bad refcon value", pKeyboardPB->pb.usbRefcon );
			pKeyboardPB->pb.usbRefcon = kUndefined + kReturnFromDriver;
			break;
	}
	
// At this point the control is returned to the system.  If a USB transaction
// has been initiated, then it will call the Complete procs
// (below) to handle the results of the transaction.
}

void KeyboardCompletionProc(USBPB *pb)
{
unsigned char	* errstring;
register usbKeyboardPBStruct *pKeyboardPB;
USBPipeState pipeState;

	pKeyboardPB = (usbKeyboardPBStruct *)(pb);
	pKeyboardPB->transDepth--;
	if (pKeyboardPB->transDepth < 0)
	{
		USBExpertFatalError(pKeyboardPB->interfaceRef, kUSBInternalErr, "\pUSBHIDKeyboardModule: transDepth < 0 (completion)", pKeyboardPB->pb.usbRefcon );
	}
	
	if (pKeyboardPB->transDepth > 1)
	{
		USBExpertFatalError(pKeyboardPB->interfaceRef, kUSBInternalErr, "\pUSBHIDKeyboardModule: transDepth > 1 (completion)", pKeyboardPB->pb.usbRefcon );
	}
	
	if(pKeyboardPB->pb.usbStatus != noErr)													// was there an error?
	{
		switch(pKeyboardPB->pb.usbRefcon & 0x0fff)											// yes, so show where the error occurred
		{
			case kSetKeyboardLEDs:  			errstring = "\pUSBHIDKeyboardModule: Error during kSetKeyboardLEDs"; break;
			case kSetProtocol:					errstring = "\pUSBHIDKeyboardModule: Error during kSetProtocol"; break;
			case kSetIdleRequest:  				errstring = "\pUSBHIDKeyboardModule: Error during kSetIdleRequest"; break;
			case kConfigureInterface:			errstring = "\pUSBHIDKeyboardModule: Error during kConfigureInterface"; break;
			case kFindPipe:  					errstring = "\pUSBHIDKeyboardModule: Error during kFindPipe"; break;
			case kReadInterruptPipe: 	 		errstring = "\pUSBHIDKeyboardModule: Error during kReadInterruptPipe"; break;
			case kClearFeature: 	 			errstring = "\pUSBHIDKeyboardModule: Error during kClearFeature"; break;
			default: 	 						errstring = "\pUSBHIDKeyboardModule: Error occurred, but state is unknown"; break;
		};
		USBExpertFatalError(pKeyboardPB->interfaceRef, pKeyboardPB->pb.usbStatus, errstring, (pKeyboardPB->pb.usbRefcon & 0x0fff));
		
		pKeyboardPB->pb.usbRefcon &= ~(kCompletionPending + kReturnFromDriver);				// set up to retry the transaction
		pKeyboardPB->pb.usbRefcon |= kRetryTransaction;
		pKeyboardPB->retryCount--;
		
		if ((!pKeyboardPB->retryCount)	|| (pKeyboardPB->pb.usbStatus == kUSBAbortedError))	// have we exhausted the retries?
		{																					// or received an abort?
			USBExpertStatus(pKeyboardPB->interfaceRef, "\pUSBHIDKeyboardModule: Pipe abort or unable to recover from error", pKeyboardPB->interfaceRef);
			pKeyboardPB->pb.usbRefcon = kReturnFromDriver;									// if so, just exit.
		}
		else																				// if it didn't abort and there's retries left, then...
		{
			if (pKeyboardPB->pipeRef)														// check if the pipe is open.
			{
				USBGetPipeStatusByReference(pKeyboardPB->pipeRef, &pipeState);				// yes, so what it's state?
				if (pipeState != kUSBActive)												// if it's not active, try to clear it.  It might be stalled...
				{
					USBExpertStatus(pKeyboardPB->interfaceRef, "\pUSBHIDKeyboardModule: Pipe is open and stalled, clearing stall...", pKeyboardPB->interfaceRef);
					USBClearPipeStallByReference(pKeyboardPB->pipeRef);
				}
			}
		}
	}
	else
	{
		pKeyboardPB->pb.usbRefcon &= ~kRetryTransaction;
		pKeyboardPB->retryCount = kKeyboardRetryCount;
	}

	if (pKeyboardPB->pb.usbRefcon & kCompletionPending)			 
	{												
		pKeyboardPB->pb.usbRefcon &= ~(kCompletionPending + kReturnFromDriver);
		switch(pKeyboardPB->pb.usbRefcon)
		{
			case kSetKeyboardLEDs:
				pKeyboardPB->pb.usbRefcon = kReturnFromDriver;
				break;
				
			case kConfigureInterface:
				pKeyboardPB->pb.usbRefcon = kSetProtocol;
				break;
				
			case kSetProtocol:
				pKeyboardPB->pb.usbRefcon = kSetIdleRequest;
				break;
				
			case kSetIdleRequest:
				pKeyboardPB->pb.usbRefcon = kFindPipe;
				break;
				
			case kFindPipe:
				pKeyboardPB->pipeRef = pKeyboardPB->pb.usbReference;
				pKeyboardPB->pb.usbRefcon = kReadInterruptPipe;
				break;
				
			case kReadInterruptPipe:
				NotifyRegisteredHIDUser(pKeyboardPB->hidDeviceType, pKeyboardPB->hidReport);
				pKeyboardPB->pb.usbRefcon = kReadInterruptPipe;
				break;
				
			case kClearFeature:
				pKeyboardPB->pb.usbRefcon = kReadInterruptPipe;
				break;
				
			default:
				USBExpertFatalError(pKeyboardPB->interfaceRef, kUSBInternalErr, "\pUSBHIDKeyboardModule - Transaction completed with bad refcon value", pKeyboardPB->pb.usbRefcon );
				pKeyboardPB->pb.usbRefcon = kUndefined + kReturnFromDriver;
				break;
		}
	}
	if (!(pKeyboardPB->pb.usbRefcon & kReturnFromDriver))
		KeyboardInitiateTransaction(pb);
}


void InterfaceEntry(UInt32 interfacenum, USBInterfaceDescriptorPtr pInterfaceDescriptor, USBDeviceDescriptorPtr pDeviceDescriptor, USBReference theInterfaceRef)
{
#pragma unused (interfacenum)

static Boolean beenThereDoneThat = false;

	if(beenThereDoneThat)
	{
		USBExpertFatalError(theInterfaceRef, kUSBInternalErr, "\pUSBHIDKeyboardModule is not reentrant", 12);
		return;
	}
	beenThereDoneThat = true;
	
	//DebugStr("\pIn Keyboard Module Interface entry routine");
	shimKeyboardPB.deviceDescriptor = *pDeviceDescriptor;				
	shimKeyboardPB.interfaceDescriptor = *pInterfaceDescriptor;				
	shimKeyboardPB.transDepth = 0;							
	shimKeyboardPB.retryCount = kKeyboardRetryCount;
	shimKeyboardPB.pSHIMInterruptRoutine = nil;
	shimKeyboardPB.pSavedInterruptRoutine = nil;
	shimKeyboardPB.interfaceRef = theInterfaceRef;
	shimKeyboardPB.pipeRef = 0;
	
	InitParamBlock(theInterfaceRef, &shimKeyboardPB.pb);
	shimKeyboardPB.pb.pbLength = sizeof(usbKeyboardPBStruct);
	shimKeyboardPB.pb.usbRefcon = 0;				

	myKeyboardPB.deviceDescriptor = *pDeviceDescriptor;				
	myKeyboardPB.interfaceDescriptor = *pInterfaceDescriptor;				
	myKeyboardPB.transDepth = 0;							
	myKeyboardPB.retryCount = kKeyboardRetryCount;
	myKeyboardPB.pSHIMInterruptRoutine = nil;
	myKeyboardPB.pSavedInterruptRoutine = nil;
	myKeyboardPB.interfaceRef = theInterfaceRef;
	myKeyboardPB.pipeRef = 0;

	myKeyboardPB.sendRawReportFlag = false;
	myKeyboardPB.hidEmulationInit = false;
	
	InitParamBlock(theInterfaceRef, &myKeyboardPB.pb);
	myKeyboardPB.pb.pbLength = sizeof(usbKeyboardPBStruct);
	myKeyboardPB.pb.usbRefcon = kConfigureInterface;					
	
	
	USBHIDControlDevice(kHIDEnableDemoMode,0);
	KeyboardInitiateTransaction(&myKeyboardPB.pb);
}



